/*
 * ----------------------------------------------------------------
 *
 * Memory and IO-Memory Access for 32 Bit bus with
 * IO-Handler registration and Translation Tables from Targets Physical
 * address (TPA) to hosts virtual Address (HVA)
 *
 * (C) 2004  Lightmaze Solutions AG
 *   Author: Jochen Karrer
 *
 *  Status:
 *	working 
 *
 *  This program is free software; you can distribute it and/or modify it
 *  under the terms of the GNU General Public License (Version 2) as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
 *
 * ----------------------------------------------------------------
 */

#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <bus.h>

/*
 * ------------------------------------
 * Hash for single IO-Ports
 * ------------------------------------
 */

IOHandler **iohandlerHash;

/*
 * ------------------------------------
 * Map for large IO-Windows 
 * ------------------------------------
 */

IOHandler **iohandlerMap;

/*
 * ------------------------------------
 * Two level Map for small IO-Windows 
 * ------------------------------------
 */

IOHandler ***iohandlerFlvlMap;
static unsigned int ioh_flvl_use_count[IOH_FLVL_SZ]={0,};

/*
 * -------------------------------------------
 * One level memory translation table vars 
 * -------------------------------------------
 */

uint8_t **mem_map_read;
uint8_t **mem_map_write;

/* 
 * -----------------------------------------
 * Two level memory translation table vars 
 * -----------------------------------------
 */

TwoLevelMMap twoLevelMMap;
InvalidateCallback *InvalidateProc;

static inline uint8_t *
twolevel_translate_r(uint32_t addr) 
{
	int index;
	uint8_t *base;
	uint8_t **slvl_map;
	index = addr >> twoLevelMMap.frst_lvl_shift; 
	if(!(slvl_map = twoLevelMMap.flvlmap_read[index])) {
		return NULL;
	}
	index = (addr & twoLevelMMap.scnd_lvl_mask) >> twoLevelMMap.scnd_lvl_shift;
	base = slvl_map[index];
	if(base) {
		return base+(addr&(twoLevelMMap.scnd_lvl_blockmask));
	} else {
		return NULL;
	}
}

static inline uint8_t *
twolevel_translate_w(uint32_t addr) 
{
	int index;
	uint8_t *base;
	uint8_t **slvl_map;
	index = addr >> twoLevelMMap.frst_lvl_shift; 
	if(!(slvl_map = twoLevelMMap.flvlmap_write[index])) {
		return NULL;
	}
	index = (addr & twoLevelMMap.scnd_lvl_mask) >> twoLevelMMap.scnd_lvl_shift;
	base = slvl_map[index];
	if(base) {
		return base+(addr&(twoLevelMMap.scnd_lvl_blockmask));
	} else {
		return NULL;
	}
}
/*
 * -----------------------------------------------
 * Bus Read
 *	Read from Memory or call an MMIO handler
 * -----------------------------------------------
 */

uint64_t
Bus_Read64(uint32_t addr) {
        uint32_t index=(addr>>MEM_MAP_SHIFT);
        uint8_t *base=mem_map_read[index];
        if(likely(base)) {
		return HMemRead64(base+(addr&(MEM_MAP_BLOCKMASK)));
        } else {
		uint8_t *taddr = twolevel_translate_r(addr);
		if(taddr) {
			return HMemRead64(taddr);
		}
		return IO_Read64(addr);
        }
}

uint32_t
Bus_Read32(uint32_t addr) {
        uint32_t index=(addr>>MEM_MAP_SHIFT);
        uint8_t *base=mem_map_read[index];
        if(likely(base)) {
		return HMemRead32(base+(addr&(MEM_MAP_BLOCKMASK)));
        } else {
		uint8_t *taddr = twolevel_translate_r(addr);
		if(taddr) {
			return HMemRead32(taddr);
		}
		return IO_Read32(addr);
        }
}

uint16_t
Bus_Read16(uint32_t addr) {
        uint32_t index=(addr>>MEM_MAP_SHIFT);
        uint8_t *base=mem_map_read[index];
        if(likely(base)) {
		return HMemRead16(base+(addr&(MEM_MAP_BLOCKMASK)));
        } else {
		uint8_t *taddr = twolevel_translate_r(addr);
		if(taddr) {
			return HMemRead16(taddr);
		}
		return IO_Read16(addr);
        }
}

uint8_t
Bus_Read8(uint32_t addr) {
        uint32_t index=(addr>>MEM_MAP_SHIFT);
        uint8_t *base=mem_map_read[index];
        if(likely(base)) {
		return HMemRead8(base+(addr&(MEM_MAP_BLOCKMASK)));
        } else {
		uint8_t *taddr = twolevel_translate_r(addr);
		if(taddr) {
			return HMemRead8(taddr);
		}
		return IO_Read8(addr);
        }
}

/*
 * --------------------------------------------
 * Bus Read
 *	Write to Memory or call an MMIO handler
 * --------------------------------------------
 */
void
Bus_Write64(uint64_t value,uint32_t addr) {
        uint32_t index=(addr>>MEM_MAP_SHIFT);
        uint8_t *base=mem_map_write[index];
        if(likely(base)) {
		HMemWrite64(value,(base+(addr&(MEM_MAP_BLOCKMASK))));
        } else {
		uint8_t *taddr = twolevel_translate_w(addr);
		if(taddr) {
			return HMemWrite64(value,taddr);
		}
		//return IO_Write64(value,addr);
		return;
        }
}

void
Bus_Write32(uint32_t value,uint32_t addr) {
        uint32_t index=(addr>>MEM_MAP_SHIFT);
        uint8_t *base=mem_map_write[index];
        if(likely(base)) {
		HMemWrite32(value,(base+(addr&(MEM_MAP_BLOCKMASK))));
        } else {
		uint8_t *taddr = twolevel_translate_w(addr);
		if(taddr) {
			return HMemWrite32(value,taddr);
		}
		return IO_Write32(value,addr);
        }
}

void
Bus_Write16(uint16_t value,uint32_t addr) {
        uint32_t index=(addr>>MEM_MAP_SHIFT);
        uint8_t *base=mem_map_write[index];
        if(likely(base)) {
		HMemWrite16(value,(base+(addr&(MEM_MAP_BLOCKMASK))));
        } else {
		uint8_t *taddr = twolevel_translate_w(addr);
		if(taddr) {
			return HMemWrite16(value,taddr);
		}
		return IO_Write16(value,addr);
        }
}

void
Bus_Write8(uint8_t value,uint32_t addr) {
        uint32_t index=(addr>>MEM_MAP_SHIFT);
        uint8_t *base=mem_map_write[index];
        if(likely(base)) {
		HMemWrite8(value,(base+(addr&(MEM_MAP_BLOCKMASK))));
        } else {
		uint8_t *taddr = twolevel_translate_w(addr);
		if(taddr) {
			return HMemWrite8(value,taddr);
		}
		return IO_Write8(value,addr);
        }
}

/*
 * --------------------------------------------
 * Generic Bus Access Functions for Transfer
 * of any block size. Mainly used for
 * non CPU bus masters.
 * --------------------------------------------
 */

void
Bus_Write(uint32_t addr,uint8_t *buf,int count) {
	while(count)  {
		Bus_Write8(*buf++,addr++);
		count--;
	}
}
void
Bus_WriteSwap32(uint32_t addr,uint8_t *buf,int count) {
	while(count)  {
		Bus_Write8(*buf++,(addr^3));
		addr++;
		count--;
	}
}
void
Bus_Read(uint8_t *buf,uint32_t addr,int count) {
	while(count)  {
		*buf=Bus_Read8(addr++);
		buf++;count--;
	}
}

void
Bus_ReadSwap32(uint8_t *buf,uint32_t addr,int count) {
	while(count)  {
		*buf=Bus_Read8(addr^3);
		addr++;
		buf++;count--;
	}
}

/*
 * ------------------------------------------------------------
 * Map/Unmap single 1MB Blocks 
 * ------------------------------------------------------------
 */
static void
Mem_MapBlock(uint32_t addr,uint8_t *host_mem,uint32_t flags) {
       	uint32_t index=(addr>>MEM_MAP_SHIFT);
	if(flags & MEM_FLAG_READABLE) {
               	mem_map_read[index]=host_mem;
	}
	if(flags & MEM_FLAG_WRITABLE) {
        	mem_map_write[index]=host_mem;
	}
}

static void
Mem_UnMapBlock(uint32_t addr) {
       	uint32_t index=(addr>>MEM_MAP_SHIFT);
        mem_map_read[index]= mem_map_write[index]=NULL;
}

/*
 * ------------------------------------------------------------
 * Map/Unmap variable sized blocks in 2-Level Translation table 
 * ------------------------------------------------------------
 */

static void
Mem_Map2LvlBlock(uint32_t addr,uint8_t *host_mem,uint32_t flags) 
{
	uint8_t **slvl_map_w;
	uint8_t **slvl_map_r;
	int index = addr >> twoLevelMMap.frst_lvl_shift;
	int sl_index = (addr & twoLevelMMap.scnd_lvl_mask) >> twoLevelMMap.scnd_lvl_shift;
	if(flags & MEM_FLAG_READABLE) {
		slvl_map_r = twoLevelMMap.flvlmap_read[index];
		if(unlikely(!slvl_map_r)) {
			slvl_map_r = twoLevelMMap.flvlmap_read[index]=malloc(sizeof(uint8_t*)*twoLevelMMap.scnd_lvl_sz);
			if(unlikely(!slvl_map_r)) {
				fprintf(stderr,"Out of memory in Map2LvlBlock\n");
				exit(4324);
			} else {
				memset(slvl_map_r,0,sizeof(uint8_t*)*twoLevelMMap.scnd_lvl_sz);
				/* assert  use count of 0 */
			}
		}  
		twoLevelMMap.flvl_map_read_use_count[index]++;
               	slvl_map_r[sl_index]=host_mem;
	}
	if(flags & MEM_FLAG_WRITABLE) {
		slvl_map_w = twoLevelMMap.flvlmap_write[index];
		if(unlikely(!slvl_map_w)) {
			slvl_map_w = twoLevelMMap.flvlmap_write[index]=malloc(sizeof(uint8_t*)*twoLevelMMap.scnd_lvl_sz);
			if(unlikely(!slvl_map_w)) {
				fprintf(stderr,"Out of memory in Map2LvlBlock\n");
				exit(4329);
			} else {
				memset(slvl_map_w,0,sizeof(uint8_t*)*twoLevelMMap.scnd_lvl_sz);
			}
		}  
		twoLevelMMap.flvl_map_write_use_count[index]++;
        	slvl_map_w[sl_index]=host_mem;
	}
		
}

static void
Mem_UnMap2LvlBlock(uint32_t addr) 
{
	uint8_t **slvl_map_w;
	uint8_t **slvl_map_r;
	int index = addr >> twoLevelMMap.frst_lvl_shift;
	int sl_index = (addr & twoLevelMMap.scnd_lvl_mask) >> twoLevelMMap.scnd_lvl_shift;
	slvl_map_r = twoLevelMMap.flvlmap_read[index];
	slvl_map_w = twoLevelMMap.flvlmap_write[index];
	if(slvl_map_r) {
        	if(slvl_map_r[sl_index]) {
        		slvl_map_r[sl_index]=NULL;
			twoLevelMMap.flvl_map_read_use_count[index]--;
			if(twoLevelMMap.flvl_map_read_use_count[index]==0) {
				free(slvl_map_r);
				slvl_map_r = twoLevelMMap.flvlmap_read[index]=NULL;
			}
		}
	}
	if(slvl_map_w) {
        	if(slvl_map_w[sl_index]) {
        		slvl_map_w[sl_index]=NULL;
			twoLevelMMap.flvl_map_write_use_count[index]--;
			if(twoLevelMMap.flvl_map_write_use_count[index]==0) {
				free(slvl_map_w);
				slvl_map_w = twoLevelMMap.flvlmap_write[index]=NULL;
			}
		}
	}
}

/*
 * ----------------------------------------------------------
 * Add a mapping to the linked list of mappings of device
 * -----------------------------------------------------------
 */
void 
Mem_AreaAddMapping(BusDevice *bdev,uint32_t base,uint32_t mapsize,uint32_t flags) 
{
	MemMapping *mapping=malloc(sizeof(MemMapping));
	if(!mapping) {
		fprintf(stderr,"Out of memory allocating MemMaping\n");
		exit(837);
	}
	mapping->next=bdev->first_mapping;
	bdev->first_mapping=mapping;
	mapping->base=base;
	mapping->mapsize = mapsize;
	mapping->flags=flags&bdev->hw_flags;

	bdev->Map(bdev->owner,base,mapping->mapsize,flags);
        /* Need to Invalidate Tlb because we cache Host Virtual Addresses */
	if(InvalidateProc) {
		InvalidateProc();
	}
}

/*
 * -------------------------------------------------------------------
 * Map/Unmap a range
 *	Forwards all decoded bits (devsize-1) to the device
 * -------------------------------------------------------------------
 */ 
void
Mem_MapRange(uint32_t base,uint8_t *start_mem,uint32_t devsize,uint32_t mapsize,uint32_t flags) {
	uint32_t count=0;
	uint32_t addr;
	uint32_t pos=base & (devsize-1);
	uint8_t *mem = start_mem+pos; 
	for(addr=base; 1 ;) {
		if((mapsize>=MEM_MAP_BLOCKSIZE) && !(addr & MEM_MAP_BLOCKMASK)) {
			Mem_MapBlock(addr,mem,flags);
			mem+=MEM_MAP_BLOCKSIZE;
			pos+=MEM_MAP_BLOCKSIZE;
			count+=MEM_MAP_BLOCKSIZE;
			addr+=MEM_MAP_BLOCKSIZE;
			mapsize-=MEM_MAP_BLOCKSIZE;
		} else if((mapsize >= twoLevelMMap.scnd_lvl_blocksize) && !(addr & twoLevelMMap.scnd_lvl_blockmask)) {
			Mem_Map2LvlBlock(addr,mem,flags);
			mem+=twoLevelMMap.scnd_lvl_blocksize;
			pos+=twoLevelMMap.scnd_lvl_blocksize;
			count+=twoLevelMMap.scnd_lvl_blocksize;
			addr+=twoLevelMMap.scnd_lvl_blocksize;
			mapsize-=twoLevelMMap.scnd_lvl_blocksize;
		} else {
			if(mapsize) {
				fprintf(stderr,"Bus: Can not map completely, rest: %08x, addr %08x"
					", twoLevelMMap.scnd_lvl_blockmask %08x\n",mapsize,addr,twoLevelMMap.scnd_lvl_blockmask);
				exit(3789);
			}
			return;
		} 
                if(pos>=devsize) {
                        pos=0;
                        mem=start_mem;
                }
                /* Special case 4GB */
                if(!count) {
                        return;
                }
        }
}

void
Mem_UnMapRange(uint32_t base,uint32_t mapsize) {
        uint32_t addr;
        uint32_t count=0;
        for(addr=base;1;) {
		if((mapsize>=MEM_MAP_BLOCKSIZE) && !(addr & MEM_MAP_BLOCKMASK)) {
			Mem_UnMapBlock(addr);
			count+=MEM_MAP_BLOCKSIZE;
			addr+=MEM_MAP_BLOCKSIZE;
			mapsize-=MEM_MAP_BLOCKSIZE;
		} else if((mapsize >= twoLevelMMap.scnd_lvl_blocksize) && !(addr & twoLevelMMap.scnd_lvl_blockmask)) {
			Mem_UnMap2LvlBlock(addr);
			count+=twoLevelMMap.scnd_lvl_blocksize;
			addr+=twoLevelMMap.scnd_lvl_blocksize;
			mapsize-=twoLevelMMap.scnd_lvl_blocksize;
		} else {
			return;
		} 
                /* Special case 4GB */
                if(!count) {
                        return;
                }
        }
}


void 
Mem_AreaDeleteMappings(BusDevice *bdev) 
{
	while(bdev->first_mapping) {
		MemMapping *mapping=bdev->first_mapping;
		twolevel_translate_r(0xfbfc);
		bdev->UnMap(bdev->owner,mapping->base,mapping->mapsize);
		bdev->first_mapping=mapping->next;
		free(mapping);
	}
        /* Need to Invalidate Tlb because we cache Host Virtual Addresses */
	if(InvalidateProc) {
		InvalidateProc();
	}
}
/*
 * ----------------------------------------------------------------------
 * Update Mappings is called by a chip emulator whenever it changes
 * its memory map by itself
 * ----------------------------------------------------------------------
 */
void 
Mem_AreaUpdateMappings(BusDevice *bdev) 
{
	MemMapping *cursor;
	/* First do a total Cleanup then map again */ 
	for(cursor=bdev->first_mapping; cursor; cursor=cursor->next) {
		bdev->UnMap(bdev->owner,cursor->base,cursor->mapsize);
	}
	for(cursor=bdev->first_mapping; cursor; cursor=cursor->next) {
		bdev->Map(bdev->owner,cursor->base,cursor->mapsize,cursor->flags);
	}
        /* Need to Invalidate Tlb because we cache Host Virtual Addresses */
	if(InvalidateProc) {
		InvalidateProc();
	}
}

/*
 * -------------------------------
 * IO-Handling
 * -------------------------------
 */
#define IOH_HASH(addr) ((addr) + ((addr)>>18))&IOH_HASH_MASK

IOHandler *
IOH_Find(uint32_t address) {
	IOHandler *cursor;
	IOHandler **slvl_map;
	uint32_t hash=IOH_HASH(address);
	uint32_t index;
	for(cursor=iohandlerHash[hash];cursor;cursor=cursor->next) {
		if(cursor->cpu_addr==address) {
			return cursor;
		}
	}
	index=(address>>IOH_MAP_SHIFT);
	if(iohandlerMap[index]) {
		return (iohandlerMap[index]);
	}
	index = (address>>IOH_FLVL_SHIFT);
	if((slvl_map = iohandlerFlvlMap[index])) {
		index = (address & IOH_SLVL_MASK) >> IOH_SLVL_SHIFT;		
		return slvl_map[index];
	} 
	return NULL;
}

IOHandler *
IOH_HashFind(uint32_t address) {
	IOHandler *cursor;
	uint32_t hash=IOH_HASH(address);
	for(cursor=iohandlerHash[hash];cursor;cursor=cursor->next) {
		if(cursor->cpu_addr==address) {
			return cursor;
		}
	}
	return NULL;
}

void
IOH_New8e(uint32_t cpu_addr,IOReadProc *readproc,IOWriteProc *writeproc,void *clientData,int byteorder) 
{
	uint32_t hash;
	IOHandler *h;
	int swap_endian;
	if(byteorder != TARGET_BYTEORDER) {
		swap_endian=1;
	} else {
		swap_endian=0;
	}
	if(IOH_HashFind(cpu_addr)) {
		fprintf(stderr,"Bug, more than one IO-Handler for address %08x\n",cpu_addr);
		exit(4324);
	}
	h=malloc(sizeof(IOHandler));
	if(!h) {
		perror("Out of memory allocating IOHandler\n");	
		exit(97);
	}
	memset(h,0,sizeof(IOHandler));
	h->cpu_addr=cpu_addr;
	h->clientData=clientData;
	h->readproc=readproc;
	h->writeproc=writeproc;
	hash=IOH_HASH(cpu_addr);
	h->next=iohandlerHash[hash];
	h->swap_endian = swap_endian;
	iohandlerHash[hash]=h;
}
void
IOH_New32e(uint32_t cpu_addr,IOReadProc *readproc,IOWriteProc *writeproc,void *clientData,int endian) {
	int i;
	for(i=0;i<4;i++) {
		IOH_New8e(cpu_addr+i,readproc,writeproc,clientData,endian); 
	}
}

void
IOH_New16(uint32_t cpu_addr,IOReadProc *readproc,IOWriteProc *writeproc,void *clientData) {
	int i;
	for(i=0;i<2;i++) {
		IOH_New8(cpu_addr+i,readproc,writeproc,clientData); 
	}
}

/*
 * -------------------------------------------------------------
 * Allocate an iohandler and initialize the structure 
 * -------------------------------------------------------------
 */
static inline IOHandler *
allocate_iohandler(uint32_t addr,IOReadProc *readproc,IOWriteProc *writeproc,int swap_endian,void *clientData) 
{
	IOHandler *h;
	h=malloc(sizeof(IOHandler));
	if(!h) {
		perror("Out of memory allocating IOHandler\n");	
		exit(97);
	}
	memset(h,0,sizeof(IOHandler));
	h->cpu_addr=addr;
	h->clientData=clientData;
	h->readproc=readproc;
	h->writeproc=writeproc;
	h->next=NULL;
	h->swap_endian = swap_endian;
	return h;
}

static inline void 
IOH_MapBlock(uint32_t addr, IOHandler *h) {
	uint32_t index=addr>>IOH_MAP_SHIFT;
	if(iohandlerMap[index]) {
		fprintf(stderr,"IO-Region %08x already allocated !\n",addr); 
		exit(5342);
	}
	iohandlerMap[index]=h;
}

static inline void
IOH_Map2LvlBlock(uint32_t addr,IOHandler *h) 
{
	IOHandler **sl_map;
	int fl_index=addr>>IOH_FLVL_SHIFT;
	int sl_index; 
	sl_map = iohandlerFlvlMap[fl_index]; 
	if(!sl_map) {
		sl_map = iohandlerFlvlMap[fl_index] = malloc(sizeof(uint8_t*)*IOH_SLVL_SZ);
		if(!sl_map) {
			fprintf(stderr,"Out of memory in IOH_Map2LvlBlock\n");
			exit(4724);
		} else {
			memset(sl_map,0,sizeof(uint8_t*)*IOH_SLVL_SZ);
			ioh_flvl_use_count[fl_index]++;
		}
	} 
	sl_index = (addr & IOH_SLVL_MASK)>>IOH_SLVL_SHIFT;
	if(!sl_map[sl_index]) {
		sl_map[sl_index] = h;
	} else {
		fprintf(stderr,"There is already a handler for IO-Region 0x%08x\n",addr);
		exit(3246);
	}
}

void
IOH_NewRegion(uint32_t addr,uint32_t length,IOReadProc *readproc,IOWriteProc *writeproc,int byteorder,void *clientData) 
{
	IOHandler *h;
	int swap_endian;
	if(byteorder != HOST_BYTEORDER) {
		swap_endian=1;
	} else {
		swap_endian=0;
	}
	while(length) {
		if(!(addr&IOH_MAP_BLOCKMASK) && (length >= IOH_MAP_BLOCKSIZE)) {
			h=allocate_iohandler(addr,readproc,writeproc,swap_endian,clientData);
			IOH_MapBlock(addr,h); 
			length-=IOH_MAP_BLOCKSIZE;
			addr+=IOH_MAP_BLOCKSIZE;
		} else if(!(addr&IOH_SLVL_BLOCKMASK) && (length >= IOH_SLVL_BLOCKSIZE)) {
			h=allocate_iohandler(addr,readproc,writeproc,swap_endian,clientData);
			IOH_Map2LvlBlock(addr,h); 
			length-=IOH_SLVL_BLOCKSIZE;
			addr+=IOH_SLVL_BLOCKSIZE;
		} else {
			fprintf(stderr,"io region not correctly aligned addr 0x%08x, len 0x%08x\n",addr,length);
			exit(34245);
		}
	}
}

static inline void
IOH_UnMapBlock(uint32_t addr) {
	int index;
	IOHandler *handler;
	index=addr>>IOH_MAP_SHIFT;
	handler=iohandlerMap[index];
	if(handler) {
		iohandlerMap[index]=NULL;
		free(handler);
	}
}
static inline void
IOH_UnMap2LvlBlock(uint32_t addr) {
	int fl_index;
	int sl_index;
	IOHandler *handler;
	IOHandler **sl_map;
	fl_index=addr>>IOH_FLVL_SHIFT;
	sl_map=iohandlerFlvlMap[fl_index];
	if(!sl_map) {
		return;
	}
	sl_index = (addr & IOH_SLVL_MASK) >> IOH_SLVL_SHIFT;
	handler = sl_map[sl_index];
	if(handler) {
		sl_map[sl_index]=NULL;
		free(handler);
		ioh_flvl_use_count[fl_index]--;
		if(ioh_flvl_use_count[fl_index]==0) {
			free(iohandlerFlvlMap[fl_index]);
			iohandlerFlvlMap[fl_index]=0;
		}
	}
}

void
IOH_DeleteRegion(uint32_t addr,uint32_t length) 
{
	//fprintf(stderr,"Delete Region %08x\n",addr);
	while(length) {
		if(!(addr&IOH_MAP_BLOCKMASK) && (length >= IOH_MAP_BLOCKSIZE)) {
			IOH_UnMapBlock(addr);
			length-=IOH_MAP_BLOCKSIZE;
			addr+=IOH_MAP_BLOCKSIZE;
		} else if(!(addr&IOH_SLVL_BLOCKMASK) && (length >= IOH_SLVL_BLOCKSIZE)) {
			IOH_UnMap2LvlBlock(addr);
			length-=IOH_SLVL_BLOCKSIZE;
			addr+=IOH_SLVL_BLOCKSIZE;

		} else {
			fprintf(stderr,"IOH_DeleteRegion: only IO regions with alignment to 0x%08x allowed\n",IOH_SLVL_BLOCKSIZE);
			return;
		}
	}

}

void
IOH_Delete8(uint32_t address) {
	IOHandler *next,*cursor,*prev=NULL;
	uint32_t hash=IOH_HASH(address);
	for(cursor=iohandlerHash[hash];cursor;prev=cursor,cursor=next) {
		next=cursor->next;
		if(cursor->cpu_addr==address) {
			if(prev) {
				prev->next=cursor->next;
			} else {
				iohandlerHash[hash]=cursor->next;
			}
			free(cursor);
			return;
		}
	}
	return;
}
void
IOH_Delete16(uint32_t address) {
	int i;
	for(i=0;i<2;i++) {
		IOH_Delete8(address+i); 
	}
}
void
IOH_Delete32(uint32_t address) {
	int i;
	for(i=0;i<4;i++) {
		IOH_Delete8(address+i); 
	}
}

/*
 * ---------------------------------------
 * Warning: 64 Bit write does not work
 * because writeproc is still 32Bit
 * ---------------------------------------
 */
void
IO_Write64(uint64_t value,uint32_t addr) { 
	IOHandler *h=IOH_Find(addr);
	if(!h || !h->writeproc) {
		fprintf(stderr,"Write: No Handler for %08x\n",addr);
		return;
	}
	h->writeproc(h->clientData,value,addr,4);
	return;
}
void
IO_Write32(uint32_t value,uint32_t addr) { 
	IOHandler *h=IOH_Find(addr);
	if(!h || !h->writeproc) {
		fprintf(stderr,"Write: No Handler for %08x\n",addr);
		return;
	}
	if(!h->swap_endian) {
		h->writeproc(h->clientData,value,addr,4);
	} else {
		h->writeproc(h->clientData,swap32(value),addr,4);
	}
	return;
}
void
IO_Write16(uint16_t value,uint32_t addr) { 
	IOHandler *h=IOH_Find(addr);
	if(!h || !h->writeproc) {
		//fprintf(stderr,"No handler for %08x\n",addr);
		return;
	}
	if(!h->swap_endian) {
		h->writeproc(h->clientData,value,addr,2);
	} else {
		h->writeproc(h->clientData,swap16(value),addr,2);
	}
	return;
}
void
IO_Write8(uint8_t value,uint32_t addr) { 
	IOHandler *h=IOH_Find(addr);
	if(!h || !h->writeproc) {
		fprintf(stderr,"No iohandler for %08x\n",addr);
		return;
	}
	h->writeproc(h->clientData,value,addr,1);
	return;
}

/*
 * ---------------------------------------------------
 * Warning: 64 Bit read does not work, because
 * readproc returns only 32 Bit
 * ---------------------------------------------------
 */
uint64_t
IO_Read64(uint32_t addr) { 
	IOHandler *h=IOH_Find(addr);
	if(!h || !h->readproc) {
		return 0;
	}
	return h->readproc(h->clientData,addr,8);
}

uint32_t
IO_Read32(uint32_t addr) { 
	IOHandler *h=IOH_Find(addr);
	if(!h || !h->readproc) {
		return 0;
	}
	if(!h->swap_endian) {
		return h->readproc(h->clientData,addr,4);
	} else {
		return swap32(h->readproc(h->clientData,addr,4));
	}
}

uint16_t
IO_Read16(uint32_t addr) { 
	IOHandler *h=IOH_Find(addr);
	if(!h || !h->readproc) {
		return 0;
	}
	if(!h->swap_endian) {
		return h->readproc(h->clientData,addr,2);
	} else {
		return swap16(h->readproc(h->clientData,addr,2));
	}
}

uint8_t
IO_Read8(uint32_t addr) { 
	IOHandler *h=IOH_Find(addr);
	if(!h || !h->readproc) {
		return 0;
	}
	return h->readproc(h->clientData,addr,1);
}

static void *
ckalloc(size_t size) {
	void *mem=malloc(size);
	if(!mem) {
		fprintf(stderr,"Out of memory\n");
		exit(423);
	}
	memset(mem,0,size);
	return mem;
}
/*
 * --------------------------------------------------------------
 * Create and Initialize memory and IO-Handler maps and Hashes
 * --------------------------------------------------------------
 */

void
Bus_Init(InvalidateCallback *invalidate,uint32_t min_memblocksize) {
	int i;
	InvalidateProc = invalidate;

	twoLevelMMap.frst_lvl_sz = (4096);
	twoLevelMMap.frst_lvl_mask =  (0xfff00000);
	twoLevelMMap.frst_lvl_shift =  (20);
	for(i=0;i<twoLevelMMap.frst_lvl_shift;i++) {
		if((min_memblocksize>>i)==1) {
			break;
		}
	}
	twoLevelMMap.scnd_lvl_shift =  i;
	twoLevelMMap.scnd_lvl_blocksize = min_memblocksize;
	twoLevelMMap.scnd_lvl_blockmask = min_memblocksize-1;
	twoLevelMMap.scnd_lvl_mask =  (~0 << i) & ~twoLevelMMap.frst_lvl_mask; 
	twoLevelMMap.scnd_lvl_sz = (1<<(twoLevelMMap.frst_lvl_shift-twoLevelMMap.scnd_lvl_shift));

	iohandlerHash=ckalloc(sizeof(IOHandler*)*IOH_HASH_SIZE);
	mem_map_read=ckalloc(sizeof(uint8_t*)*MEM_MAP_ENTRIES);
	mem_map_write=ckalloc(sizeof(uint8_t*)*MEM_MAP_ENTRIES);

	twoLevelMMap.flvlmap_read=ckalloc(sizeof(uint8_t*)*twoLevelMMap.frst_lvl_sz);
	twoLevelMMap.flvlmap_write=ckalloc(sizeof(uint8_t*)*twoLevelMMap.frst_lvl_sz);
	twoLevelMMap.flvl_map_read_use_count=ckalloc(sizeof(*twoLevelMMap.flvl_map_read_use_count)*twoLevelMMap.frst_lvl_sz);
	twoLevelMMap.flvl_map_write_use_count=ckalloc(sizeof(*twoLevelMMap.flvl_map_read_use_count)*twoLevelMMap.frst_lvl_sz);

	iohandlerMap=ckalloc(sizeof(IOHandler *)*IOH_MAP_ENTRIES);
	iohandlerFlvlMap=ckalloc(sizeof(IOHandler **)*IOH_FLVL_SZ);
	fprintf(stderr,"MemMap and IO-Handler Hash initialized\n");
}
