/*
 * ------------------------------------------------------------------------
 * DRAM Emulation
 *
 * (C) 2004  Lightmaze Solutions AG
 *   Author: Jochen Karrer
 *
 *  State:
 *	Working but no timing checks. It works even when the 
 *	real CPU will crash because of wrong Number of Row/Column addresses 
 *	or wrong timings.
 *
 *  This program is free software; you can distribute it and/or modify it
 *  under the terms of the GNU General Public License (Version 2) as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
 * ------------------------------------------------------------------------
 *
 */
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>

#include "bus.h"
#include "configfile.h"
#include "dram.h"

/* all times in nanoseconds, all clocks in cycles */
typedef struct DRamTiming {
	int tRP; 	// Precharge Command Period		
	int tRAS;	// Active to Precharge 
	int tSREX;	// Self refresh exit time 
	int tAPR;	// Last data out to active time
 	int tDAL;	// Data in to active time
	int tWR;	// Write recovery time
	int tRC;	// Active to active cmd period
	int tRFC;	// Auto Refresh periaod and refresh to active
	int tXSR;	// exit self refresh to active cmd
	int tRRD;	// active a to active b
	int tMRD;	// Load mode to active command
	int cRAS;	// RAS latency clocks
 	int cCAS;	// CAS latency clocks
} DRamTiming;

typedef struct DRamType {
	char *name;
	uint32_t size_mbits;
	DRamTiming timing;
} DRamType;

DRamType dram_types[] = {
	{ 
		name: "K4S2816F32-TC75",	
		size_mbits: 128,
		timing: {
			tRP: 	20,
			tRAS:	100, 
			tSREX:	10,
			tAPR:	10, /* ??? */
			tDAL:   40,
			tWR:	20, /* tRDL */
			tRC:	65,
			tRFC:	70, /* Infineon */
			tXSR:	10, /* ????? */
			tRRD:	15,
			tMRD:	10,
			cRAS:   2,		
			cCAS:   2
		}
	}
	
};

typedef struct DRam {
	uint8_t *host_mem;
	uint32_t  size;
	uint32_t flags;
	BusDevice bdev;
} DRam;

static void
DRam_Map(void *module_owner,uint32_t base,uint32_t mapsize,uint32_t flags) {
	DRam *dram = module_owner;
	flags &= MEM_FLAG_READABLE|MEM_FLAG_WRITABLE;
	Mem_MapRange(base,dram->host_mem,dram->size,mapsize,flags);
}

static void
DRam_UnMap(void *module_owner,uint32_t base,uint32_t mapsize) {
	Mem_UnMapRange(base,mapsize); 
}

static uint32_t
parse_memsize (char *str)
{
        uint32_t size;
        char c;
        if(sscanf(str,"%d",&size)!=1) {
                return 0;
        }
        if(sscanf(str,"%d%c",&size,&c)==1) {
                return size;
        }
        switch(tolower(c)) {
                case 'm':
                        return size*1024*1024;
                case 'k':
                        return size*1024;
        }
        return 0;
}

/*
 * --------------------
 * DRAM New
 * --------------------
 */
BusDevice *
DRam_New(char *dram_name) {
	char *sizestr;
	uint32_t size=0;
	DRam *dram;
	sizestr=Config_ReadVar(dram_name,"size");
	if(sizestr) {
		size=parse_memsize(sizestr);
		if(size==0) {
			fprintf(stderr,"DRAM bank \"%s\" not present\n",dram_name);
			return NULL;
		}
	} else {
		fprintf(stderr,"DRAM bank \"%s\" not present\n",dram_name);
		return NULL;
	}
	dram = malloc(sizeof(DRam));
	if(!dram) {
		fprintf(stderr,"Out of memory\n");
		exit(556);
	}
	memset(dram,0,sizeof(DRam));
	dram->host_mem=malloc(size);
	if(!dram->host_mem) {
		fprintf(stderr,"Can not allocate memory for DRAM emulation\n");
		exit(546);
	}
	memset(dram->host_mem,0xff,size);
	dram->size=size;
	dram->bdev.first_mapping=NULL;
	dram->bdev.Map=DRam_Map;
	dram->bdev.UnMap=DRam_UnMap;
	dram->bdev.owner=dram;
	dram->bdev.hw_flags=MEM_FLAG_WRITABLE|MEM_FLAG_READABLE;
	fprintf(stderr,"DRAM bank \"%s\" with  size %ukB\n",dram_name,size/1024);
	return &dram->bdev;
}
