/*
 * ------------------------------------------------------------------------------
 *
 * Emulation of PCF8563 I2C-Realtime Clock 
 *
 * (C) 2004  Lightmaze Solutions AG
 *   Author: Jochen Karrer
 *
 * State: Time is readable, writes are ignored Timezones and daylightsavingtime 
 *        is not correct, Alarms are missing
 *
 *  This program is free software; you can distribute it and/or modify it
 *  under the terms of the GNU General Public License (Version 2) as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
 * ------------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>
#include "i2c.h"
#include "pcf8563.h"

#if 0
#define dprintf(x...) { fprintf(stderr,x); }
#else
#define dprintf(x...)
#endif

#define PCF_STATE_ADDR (0)
#define PCF_STATE_DATA  (2)

struct PCF8563 {
	I2C_Slave i2c_slave;
	uint16_t reg_address;
	int state;
	uint8_t data[16];
};

/*
 * ------------------------------------
 * PCF8563 Write state machine 
 * ------------------------------------
 */
static int 
pcf8563_write(void *dev,uint8_t data) {
	PCF8563 *pcf = dev;
	if(pcf->state==PCF_STATE_ADDR) {
		dprintf("PCF8563 Addr 0x%02x\n",data);
		pcf->reg_address = data & 0xf;
		pcf->state = PCF_STATE_DATA;
	} else if(pcf->state==PCF_STATE_DATA) {
		dprintf("PCF8563 Write 0x%02x to %04x\n",data,pcf->reg_address);
		pcf->data[pcf->reg_address]=data;
		pcf->reg_address=((pcf->reg_address+1)&0xf) | (pcf->reg_address & ~0xf);
	}
	return I2C_ACK;
};

static void 
pcf8563_read(void *dev,uint8_t *data,int ack) 
{
	PCF8563 *pcf = dev;
	*data=pcf->data[pcf->reg_address];
	dprintf("PCF8563 read 0x%02x from %04x\n",*data,pcf->reg_address);
	pcf->reg_address=(pcf->reg_address+1)&0xf;
};

static inline unsigned char 
bcd_to_i(unsigned char b) {
        return (b&0xf)+10*((b>>4)&0xf);
}
static inline unsigned char 
i_to_bcd(unsigned char b) {
        return (b/10)*16+(b-10*(b/10));
}

#define RTC_SECONDS  2
#define RTC_MINUTES  3
#define RTC_HOURS    4
#define RTC_DAYS     5
#define RTC_WEEKDAYS 6
#define RTC_MONTHS   7
#define RTC_YEARS    8

/*
 * ----------------------------------------------------------------
 * read_systemtime
 * 	Read new time from system and convert to PCF8563 format 
 *	Called on I2C-Start with direction READ	
 * ----------------------------------------------------------------
 */
static void
read_systemtime(PCF8563 *pcf) {
	struct timeval tv;
	time_t time;
	struct tm *tm;
	gettimeofday(&tv,NULL);
	time = tv.tv_sec;
	tm = localtime(&time);
	pcf->data[RTC_MONTHS] = i_to_bcd(tm->tm_mon+1);
	pcf->data[RTC_DAYS] = i_to_bcd(tm->tm_mday);
	pcf->data[RTC_YEARS] =  i_to_bcd(tm->tm_year - 100);
	pcf->data[RTC_HOURS] =  i_to_bcd(tm->tm_hour);
	pcf->data[RTC_MINUTES] =  i_to_bcd(tm->tm_min);
	pcf->data[RTC_SECONDS] =  i_to_bcd(tm->tm_sec);
}
static int
pcf8563_start(void *dev,int i2c_addr,int operation) {
	PCF8563 *pcf = dev;
	dprintf("pcf8563 start\n");
	pcf->state = PCF_STATE_ADDR;
	if(operation == I2C_READ) {
		read_systemtime(pcf); 
	}
	return I2C_ACK;
}

static void 
pcf8563_stop(void *dev) {
	PCF8563 *pcf = dev;
	dprintf("pcf8563 stop\n");
	pcf->state =  PCF_STATE_ADDR; 
}

/*
 * -----------------------------------------------
 * pcf8563_ops 
 *	I2C-Operations provided by the PCF8563
 * -----------------------------------------------
 */

static I2C_SlaveOps pcf8563_ops = {
	.start = pcf8563_start,
	.stop =  pcf8563_stop,
	.read =  pcf8563_read,	
	.write = pcf8563_write	
};

I2C_Slave *
PCF8563_New(char *name) {
	PCF8563 *pcf = malloc(sizeof(PCF8563)); 
	I2C_Slave *i2c_slave;
	if(!pcf) {
		fprintf(stderr,"Out of memory\n");
		exit(8152);
	}
	memset(pcf,0,sizeof(PCF8563));
	i2c_slave = &pcf->i2c_slave;
	i2c_slave->devops = &pcf8563_ops; 
	i2c_slave->dev = pcf;
	i2c_slave->speed = I2C_SPEED_FAST;
	fprintf(stderr,"PCF8563 Real Time Clock \"%s\" created\n",name);
	return i2c_slave;
}
